defmodule Diffuser.GeneratorTest do
  use Diffuser.DataCase

  alias Diffuser.Generator
  alias Diffuser.Test.FileHelper
  alias Diffuser.Repo

  describe "prompt_requests" do
    alias Diffuser.Generator.PromptRequest

    import Diffuser.GeneratorFixtures

    @invalid_attrs %{prompt: nil}

    test "list_prompt_requests/0 returns all prompt_requests" do
      prompt_request = prompt_request_fixture()
      assert Generator.list_prompt_requests() == [prompt_request]
    end

    test "get_prompt_request!/1 returns the prompt_request with given id" do
      prompt_request = prompt_request_fixture()
      assert Generator.get_prompt_request!(prompt_request.id) == prompt_request
    end

    test "create_prompt_request/1 with valid data creates a prompt_request" do
      valid_attrs = %{prompt: "some prompt"}

      assert {:ok, %PromptRequest{} = prompt_request} =
               Generator.create_prompt_request(valid_attrs)

      assert prompt_request.prompt == "some prompt"
    end

    test "create_prompt_request/1 with invalid data returns error changeset" do
      assert {:error, %Ecto.Changeset{}} = Generator.create_prompt_request(@invalid_attrs)
    end

    test "update_prompt_request/2 with valid data updates the prompt_request" do
      prompt_request = prompt_request_fixture()

      update_attrs = %{
        prompt: "some updated prompt"
      }

      assert {:ok, %PromptRequest{} = prompt_request} =
               Generator.update_prompt_request(prompt_request, update_attrs)

      assert prompt_request.prompt == "some updated prompt"
    end

    test "update_prompt_request/2 with invalid data returns error changeset" do
      prompt_request = prompt_request_fixture()

      assert {:error, %Ecto.Changeset{}} =
               Generator.update_prompt_request(prompt_request, @invalid_attrs)

      assert prompt_request == Generator.get_prompt_request!(prompt_request.id)
    end

    test "delete_prompt_request/1 deletes the prompt_request" do
      prompt_request = prompt_request_fixture()
      assert {:ok, %PromptRequest{}} = Generator.delete_prompt_request(prompt_request)
      assert_raise Ecto.NoResultsError, fn -> Generator.get_prompt_request!(prompt_request.id) end
    end

    test "change_prompt_request/1 returns a prompt_request changeset" do
      prompt_request = prompt_request_fixture()
      assert %Ecto.Changeset{} = Generator.change_prompt_request(prompt_request)
    end

    test "begin_prompt_request/1 returns pid of a prompt request generator" do
      prompt_request = prompt_request_fixture()
      assert {:ok, _pid} = Generator.begin_prompt_request(prompt_request)
    end

    test "create_prompt_request_results/2 properly associates images to prompt request" do
      %{id: prompt_request_id} = prompt_request_fixture()

      images = [
        %{file_name: "cat.jpg", binary: FileHelper.binary_image()},
        %{file_name: "dog.jpg", binary: FileHelper.binary_image()},
        %{file_name: "horse.jpg", binary: FileHelper.binary_image()}
      ]

      Generator.create_prompt_request_results(prompt_request_id, images)

      %PromptRequest{images: images} = Generator.get_prompt_request!(prompt_request_id)
      assert images |> length() > 0
      # assert images |> Enum.all?
    end
  end
end
